package j1.tutorial.javac.analysis.api;

import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.AnnotationValue;
import javax.lang.model.element.Element;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.TypeElement;
import javax.lang.model.element.VariableElement;
import javax.lang.model.type.TypeMirror;
import javax.lang.model.util.ElementFilter;
import javax.lang.model.util.Elements;

public class Utilities {

    private Utilities() {
    }

    public static AnnotationMirror findAppliedAnnotation(AnalysisContext context, Element forElement, String annotationFQN) {
        return findAppliedAnnotation(forElement, annotationFQN);
    }
    
    public static AnnotationMirror findAppliedAnnotation(ProcessingEnvironment context, Element forElement, String annotationFQN) {
        return findAppliedAnnotation(forElement, annotationFQN);
    }
    
    private static AnnotationMirror findAppliedAnnotation(Element forElement, String annotationFQN) {
        while (forElement != null) {
            AnnotationMirror result = findAttachedAnnotation(forElement, annotationFQN);
            
            if (result != null) return result;
            
            forElement = forElement.getEnclosingElement();
        }

        return null;
    }
    
    public static AnnotationMirror findAttachedAnnotation(ProcessingEnvironment context, Element forElement, String annotationFQN) {
        return findAttachedAnnotation(forElement, annotationFQN);
    }
    
    public static AnnotationMirror findAttachedAnnotation(AnalysisContext context, Element forElement, String annotationFQN) {
        return findAttachedAnnotation(forElement, annotationFQN);
    }
    
    private static AnnotationMirror findAttachedAnnotation(Element forElement, String annotationFQN) {
        for (AnnotationMirror am : forElement.getAnnotationMirrors()) {
            if (((TypeElement) am.getAnnotationType().asElement()).getQualifiedName().contentEquals(annotationFQN)) {
                return am;
            }
        }

        return null;
    }
    
    public static <T> T getAttributeValue(ProcessingEnvironment context, AnnotationMirror annotation, String attribute, Class<T> clazz) {
        return getAttributeValue(context.getElementUtils(), annotation, attribute, clazz);
    }
    
    public static <T> T getAttributeValue(AnalysisContext context, AnnotationMirror annotation, String attribute, Class<T> clazz) {
        return getAttributeValue(context.getElementUtils(), annotation, attribute, clazz);
    }
    
    private static <T> T getAttributeValue(Elements elements, AnnotationMirror annotation, String attribute, Class<T> clazz) {
        if (clazz.isArray()) {
            Iterable<?> attributes = getAttributeValueInternal(elements, annotation, attribute, Iterable.class);
            Collection<Object> coll = new ArrayList<>();

            for (Object internal : attributes) {
                coll.add(clazz.getComponentType().cast(AnnotationValue.class.cast(internal).getValue()));
            }

            return clazz.cast(coll.toArray((Object[]) Array.newInstance(clazz.getComponentType(), 0)));
        } else {
            return getAttributeValueInternal(elements, annotation, attribute, clazz);
        }
    }

    private static <T> T getAttributeValueInternal(Elements elements, AnnotationMirror annotation, String attribute, Class<T> clazz) {
        for (Entry<? extends ExecutableElement, ? extends AnnotationValue> e : elements.getElementValuesWithDefaults(annotation).entrySet()) {
            if (e.getKey().getSimpleName().contentEquals(attribute)) {
                Object value = e.getValue().getValue();

                if (clazz.isAssignableFrom(value.getClass())) {
                    return clazz.cast(value);
                }

                return null;
            }
        }
        
        return null;
    }
    
    public static String toFQN(TypeMirror type) {
        return type.toString(); //XXX: should not depend on toString!
    }
    
    private static final Pattern SPLIT = Pattern.compile("(.+)\\.([^.]+)\\((.*)\\)");
    
    public static ExecutableElement resolveMethod(AnalysisContext ctx, String name) {
        Matcher m = SPLIT.matcher(name);

        if (!m.matches()) {
            throw new IllegalArgumentException();
        }

        String className = m.group(1);
        String methodName = m.group(2);
        String paramsSpec = m.group(3);

        TypeElement te = ctx.getElementUtils().getTypeElement(className);

        if (te == null) {
            return null;
        }

        String[] paramList = paramsSpec.split(",");
        List<TypeMirror> params = new LinkedList<TypeMirror>();

        for (String t : paramList) {
            TypeElement param = ctx.getElementUtils().getTypeElement(t);
            params.add(ctx.getTypeUtils().erasure(param.asType()));
        }

        for (ExecutableElement ee : ElementFilter.methodsIn(te.getEnclosedElements())) {
            if (!methodName.equals(ee.getSimpleName().toString()) || ee.getParameters().size() != params.size()) {
                continue;
            }
            
            Iterator<TypeMirror> designed = params.iterator();
            boolean found = true;

            for (VariableElement param : ee.getParameters()) {
                if (!ctx.getTypeUtils().isSameType(ctx.getTypeUtils().erasure(param.asType()), designed.next())) {
                    found = false;
                    break;
                }
            }

            if (found) {
                return ee;
            }
        }
        
        return null;
    }

}
