package j1.tutorial.javac.analysis.api;

import com.sun.source.tree.CompilationUnitTree;
import com.sun.source.util.JavacTask;
import java.io.File;
import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import javax.annotation.processing.Processor;
import javax.tools.Diagnostic;
import javax.tools.DiagnosticCollector;
import javax.tools.JavaCompiler;
import javax.tools.JavaFileObject;
import javax.tools.SimpleJavaFileObject;
import javax.tools.ToolProvider;
import static org.testng.Assert.*;

public class RuleTest {

    private RuleTest() {
    }
    
    public static RuleTest create() {
        return new RuleTest();
    }
    
    public RuleTest input(String code) {
        input("test/Test.java", code);
        return this;
    }
    
    public RuleTest input(String filename, String code) {
        files.add(new FileDescription(filename, code));
        return this;
    }
    
    public RuleResult run(final Class<? extends Rule> r) throws Exception {
        return run(new AnalyzerImpl() {
            @Override public void analyze(JavacTask javac, CompilationUnitTree cut) throws Exception {
                r.newInstance().analyseFile(new AnalysisContext(javac), cut);
            }
        }, "-proc:none");
    }
    
    public RuleResult runProcessor(final Class<? extends Processor> r) throws Exception {
        File processorJar = new File(r.getProtectionDomain().getCodeSource().getLocation().toURI());
        File apiJar = new File(Friend.class.getProtectionDomain().getCodeSource().getLocation().toURI());
        return run(null, "-processorpath", processorJar.getAbsolutePath() + File.pathSeparatorChar + apiJar.getAbsolutePath(), "-processor", r.getName(), "-proc:only");
    }
    
    private RuleResult run(AnalyzerImpl analyzer, String... extraOptions) throws Exception {
        final String bootPath = System.getProperty("sun.boot.class.path"); //NOI18N
        final JavaCompiler tool = ToolProvider.getSystemJavaCompiler();
        assert tool != null;

        List<JavaFileObject> fos = new ArrayList<>(files.size());
        
        for (FileDescription fd : files) {
            fos.add(new MyFileObject(fd.fileName, fd.code));
        }
        
        File apiJar = new File(Friend.class.getProtectionDomain().getCodeSource().getLocation().toURI());
        DiagnosticCollector<JavaFileObject> dc = new DiagnosticCollector<>();
        List<String> options = new ArrayList<>(Arrays.asList("-bootclasspath",  bootPath, "-Xjcov", "-classpath", apiJar.getAbsolutePath()));
        options.addAll(Arrays.asList(extraOptions));
        final JavacTask ct = (JavacTask) tool.getTask(null, null, dc, options, null, fos);
        Iterable<? extends CompilationUnitTree> cuts = ct.parse();
        
        ct.analyze();
        
        if (analyzer != null) {
            for (CompilationUnitTree cut : cuts) {
                analyzer.analyze(ct, cut);
            }
        }
        
        List<String> warnings = new ArrayList<>();
        
        for (Diagnostic<? extends JavaFileObject> d : dc.getDiagnostics()) {
            if (   !"compiler.err.proc.messager".equals(d.getCode())
                && !"compiler.warn.proc.messager".equals(d.getCode())) continue;
            if (d.getSource() == null) continue;
            warnings.add(d.getSource().getName() + ":" + d.getStartPosition() + "-" + d.getEndPosition() + ":" + d.getMessage(Locale.ENGLISH));
        }
        
        return new RuleResult(warnings);
    }
    
    private final List<FileDescription> files = new ArrayList<>();
    
    private static final class FileDescription {
        private final String fileName;
        private final String code;

        public FileDescription(String fileName, String code) {
            this.fileName = fileName;
            this.code = code;
        }

    }
    
    private static class MyFileObject extends SimpleJavaFileObject {
        private String text;
        public MyFileObject(String text) {
            this("Test", text);
        }
        public MyFileObject(String name, String text) {
            super(URI.create("myfo:/" + name), JavaFileObject.Kind.SOURCE);
            this.text = text;
        }
        @Override
        public CharSequence getCharContent(boolean ignoreEncodingErrors) {
            return text;
        }
    }

    public static class RuleResult {
        private final List<String> warnings;

        private RuleResult(List<String> warnings) {
            this.warnings = warnings;
        }
        
        public void assertWarnings(String... expectedWarnings) {
            assertEquals(warnings, Arrays.asList(expectedWarnings), warnings.toString());
        }
    }

    private static interface AnalyzerImpl {

        public void analyze(JavacTask javac, CompilationUnitTree cut) throws Exception;
        
    }
}
